/* SmoothWall install program.
 *
 * By The SmoothWall Team.
 * 
 * This program is distributed under the terms of the GNU General Public
 * Licence.  See the file COPYING for details.
 *
 * filename: main.c
 * Contains main entry point, and misc functions. */

#include "install.h"

#define CDROM_INSTALL 0
#define URL_INSTALL 1

FILE *flog = NULL;
char *log;

int main(int argc, char *argv[])
{
	char letter;		/* Drive letter holder. */
	struct devparams hdparams, cdromparams;	/* Params for CDROM and HD */
	char mountcommand[STRING_SIZE]; /* Mount command string (for /cdrom) */
	int cdmounted;			/* Loop flag for inserting a cd. */ 
	int rc;
	char commandstring[STRING_SIZE];
	char *installtypes[] = { "CDROM", "FTP or HTTP", NULL };
	int installtype = CDROM_INSTALL; 
	char *insertmessage, *insertdevnode;
	struct networkparams networkparams;
	char tarballfilename[STRING_SIZE];
	
	memset(&hdparams, 0, sizeof(struct devparams));
	memset(&cdromparams, 0, sizeof(struct devparams));
	memset(&networkparams, 0, sizeof(struct networkparams));
	
	/* Log file/terminal stuff. */
	if (argc >= 2)
	{		
		if (!(flog = fopen(argv[1], "w+")))
		{
			printf("Couldn't open log terminal\n");
			return 0;
		}
	}
	else
	{
		printf("Log terminal not given\n");
		return 0;
	}
	
	log = argv[1];
	
	fprintf(flog, "Install program started.\n");
	
	newtInit();
	newtCls();
	
	newtDrawRootText(7, 0, "SmoothWall - Lawrence Manning, Richard Morrell www.smoothwall.org");
	newtPushHelpLine("             <Tab>/<Alt-Tab> between elements   |  <Space> selects");		

	newtWinMessage(TITLE, "Ok",
		"Welcome to the SmoothWall installation program. Please visit the SmoothWall " \
		"homepage at http://www.smoothwall.org.  Selecting Cancel on " \
		"any of the following screens will reboot the computer.");
						
	/* Get device letter for the IDE HD.  This has to succeed. */
	if (!(letter = findidetype(IDE_HD)))
	{
		errorbox("No IDE harddisk found.");
		goto EXIT;
	}
	
	fprintf(flog, "Found Harddisk.\n");
	
	/* Make the hdparms struct and print the contents. */
	snprintf(hdparams.devnode, STRING_SIZE, "/dev/hd%c", letter);
	hdparams.module = 0;

	rc = newtWinMenu("Select installation media", "SmoothWall can be installed " \
		"from multiple sources.  The simplest is to use the machines CDROM drive. " \
		"If the computer lacks a drive, you may install via another machine on the " \
		"LAN which has the installation files available via FTP or HTTP. In this " \
		"case the network driver diskette will be required.",
		50, 5, 5, 6, installtypes, &installtype, "Ok", "Cancel", NULL);
	
	if (rc == 2)
		goto EXIT;	
	
	fprintf(flog, "Instalation type: %d\n", installtype);

	if (installtype == CDROM_INSTALL)
	{	
		fprintf(flog, "Looking for IDE CDROM...\n");
	
		/* First look for an IDE CDROM. */
		if (!(letter = findidetype(IDE_CDROM)))
		{
			/* If it fails, go the CDROM menu. */
			fprintf(flog, "No IDE CDROM found.\n");
			if (!(cdrommenu(&cdromparams)))
			{
				/* Didn't work, oops. */
				errorbox("No CDROM found.");
				goto EXIT;
			}
		}
		else
		{
			snprintf(cdromparams.devnode, STRING_SIZE, "/dev/hd%c", letter);
			cdromparams.module = 0;
		}
		
		insertmessage = "Please insert the SmoothWall CD in the CDROM drive.";
		insertdevnode = cdromparams.devnode;
	}
	else
	{
		insertmessage = "Please insert the SmoothWall driver diskette in the " \
			"floppy drive.";
		insertdevnode = "/dev/fd0";
	}

	/* Try to mount /cdrom in a loop. */
	cdmounted = 0;
	snprintf(mountcommand, STRING_SIZE, "/bin/mount %s /cdrom", insertdevnode);
	while (!cdmounted)
	{
		rc = newtWinChoice(TITLE, "Ok", "Cancel", insertmessage);
		if (rc != 1)
		{
			errorbox("Installation canceled.");
			goto EXIT;
		}
		if (!(mysystem(mountcommand)))
			cdmounted = 1;
		else
			fprintf(flog, "Mount command failed.\n");
	}

	fprintf(flog, "HD: devnode=%s, module=%d, modulename=%s, options=%s\n",
		hdparams.devnode, hdparams.module, hdparams.modulename, hdparams.options);
	fprintf(flog, "CDROM: devnode=%s, module=%d, modulename=%s, options=%s\n",
		cdromparams.devnode, cdromparams.module, cdromparams.modulename, cdromparams.options);
	
	rc = newtWinChoice(TITLE, "Ok", "Cancel", "The installation program will now prepare the " \
		"IDE harddisk on %s. First the disk will be partitioned, and then the partitions " \
		"will have a filesystem put on them.", hdparams.devnode);
	if (rc != 1)
		goto EXIT;

	/* Partition, mkswp, mkfs. */
	snprintf(commandstring, STRING_SIZE, "/cdrom/bin/sfdisk -uM %s < /cdrom/partitiontable", hdparams.devnode);
	if (runcommandwithstatus(commandstring, "Partitioning disk..."))
	{
		errorbox("Unable to partition the disk.");
		goto EXIT;
	}
	snprintf(commandstring, STRING_SIZE, "/cdrom/bin/mkswap %s1", hdparams.devnode);
	if (runcommandwithstatus(commandstring, "Making swap partition..."))
	{
		errorbox("Unable to make swapspace.");
		goto EXIT;
	}
	snprintf(commandstring, STRING_SIZE, "/cdrom/bin/mke2fs %s2", hdparams.devnode);	
	if (runcommandwithstatus(commandstring, "Making filesystem..."))
	{
		errorbox("Unable to make the filesystem.");
		goto EXIT;
	}
	/* Mount harddisk. */
	snprintf(commandstring, STRING_SIZE, "/sbin/mount %s2 /harddisk", hdparams.devnode);
	if (runcommandwithstatus(commandstring, "Mounting root partition..."))
	{
		errorbox("Unable to mount the root filesystem.");
		goto EXIT;
	}
	
	/* Network driver and params. */
	if (!(networkmenu(&networkparams)))
	{
		errorbox("Network setup failed.");
		goto EXIT;
	}

	/* Either use traball from cdrom or download. */
	if (installtype == CDROM_INSTALL)
		strncpy(tarballfilename, "/cdrom/smoothwall.tgz", STRING_SIZE);
	else
	{
		if (!(downloadtarball()))
		{
			errorbox("No tarball downloaded.");
			goto EXIT;
		}
		strncpy(tarballfilename, "/harddisk/smoothwall.tgz", STRING_SIZE);
	}
	
	/* unpack... */
	snprintf(commandstring, STRING_SIZE, 
		"/cdrom/bin/gzip -d -c %s | /cdrom/bin/tar -C /harddisk -xvf -",
		tarballfilename);
	if (runcommandwithprogress(30, 4, TITLE, commandstring, 7129,
		"Installing files..."))
	{
		errorbox("Unable to install files.");
		goto EXIT;
	}
	
	/* Remove temp tarball if we downloaded. */
	if (installtype == URL_INSTALL)
	{
		if (unlink("/harddisk/smoothwall.tgz"))
		{
			errorbox("Unable to remove temporary downloaded files");
			goto EXIT;
		}
	}
	
	if (!(writeconfigs(&hdparams, &networkparams)))
	{
		errorbox("Error writing configuration information.");
		goto EXIT;
	}
	
	if (runcommandwithstatus("/harddisk/sbin/lilo -r /harddisk",
		"Installing LILO..."))
	{
		errorbox("Unable to install LILO.");
		goto EXIT;
	}
	
	if (mysystem("/sbin/umount /harddisk"))
	{
		errorbox("Unable to unmount harddisk.");
		goto EXIT;
	}
	if (mysystem("/sbin/umount /cdrom"))
	{
		errorbox("Unable to unmount the CDROM/floppydisk.");
		goto EXIT;
	}

	if (installtype == CDROM_INSTALL)
	{
		snprintf(commandstring, STRING_SIZE, "/bin/eject %s", cdromparams.devnode);
		if (mysystem(commandstring))
		{
			errorbox("Unable to eject the CDROM.");
			goto EXIT;
		}
	}
	
	newtWinMessage("Congratulations!", "Ok",
		"SmoothWall was successfully installed. Please remove any floppy disks or " \
		"CDROMs in the computer. After installation, you should configure " \
		"SmoothWall by pointing your web browser at this machine. Remember to set " \
		"a password for the SmoothWall 'dial' user, if you wish non SmoothWall "
		"'admin' users to be able to control the link.");
				
EXIT:	
	fprintf(flog, "Install program ended.\n");	
	fflush(flog);
	fclose(flog);
		
	newtWinMessage(TITLE, "Ok",	"Press Ok to reboot.");
	newtFinished();

	reboot();

	return 0;
}

/* reboot().  Waits for a key and then reboots. */
void reboot(void)
{
	mysystem("/sbin/init 6");
}

/* stripnl().  Replaces \n with \0 */
void stripnl(char *s)
{
	char *t = strchr(s, '\n');
	if (t) *t = '\0';
}

/* Little wrapper. */
int mysystem(char *command)
{
	char mycommand[STRING_SIZE];
	
	snprintf(mycommand, STRING_SIZE, "%s >>%s 2>>%s", command, log, log);
	fprintf(flog, "Running command: %s\n", command);
	return system(mycommand);
}

void errorbox(char *message)
{
	newtWinMessage("ERROR", "Ok", message);
}

void statuswindow(int width, int height, char *title, char *text, ...)
{
	newtComponent t, f;
	char *buf = NULL;
	int size = 0;
	int i = 0;
	va_list args;

	va_start(args, text);

	do {
		size += 1000;
		if (buf) free(buf);
		buf = malloc(size);
		i = vsnprintf(buf, size, text, args);
	} while (i == size);

	va_end(args);

	newtCenteredWindow(width, height, title);

	t = newtTextbox(1, 1, width - 2, height - 2, NEWT_TEXTBOX_WRAP);
	newtTextboxSetText(t, buf);
	f = newtForm(NULL, NULL, 0);

	free(buf);

	newtFormAddComponent(f, t);

	newtDrawForm(f);
	newtRefresh();
	newtFormDestroy(f);
}

int runcommandwithstatus(char *command, char *message)
{
	int rc;
	
	statuswindow(60, 4, TITLE, message);
	rc = mysystem(command);
	newtPopWindow();
	
	return rc;
}

/* This one borrowed from redhat installer. */
int runcommandwithprogress(int width, int height, char *title, char *command,
	int lines, char *text, ...)
{
	newtComponent t, f, s;
	char *buf = NULL;
	int size = 0;
	int i = 0;
	va_list args;
	int rc = 0;
	FILE *p;
	char buffer[STRING_SIZE];
	int progress = 0;

	va_start(args, text);

	do {
		size += 1000;
		if (buf) free(buf);
		buf = malloc(size);
		i = vsnprintf(buf, size, text, args);
	} while (i == size);

	va_end(args);

	newtCenteredWindow(width, height, title);

	t = newtTextbox(1, 1, width - 2, height - 2, NEWT_TEXTBOX_WRAP);
	newtTextboxSetText(t, buf);
	f = newtForm(NULL, NULL, 0);

	free(buf);

	newtFormAddComponent(f, t);
	
	s = newtScale(1, 3, width - 2, lines);
	newtScaleSet(s, progress);
	
	newtFormAddComponent(f, s);

	newtDrawForm(f);
	newtRefresh();
	
	if (!(p = popen(command, "r")))
	{
		rc = 1;
		goto EXIT;
	}
	setvbuf(p, NULL, _IOLBF, 255);
	
	while (fgets(buffer, STRING_SIZE, p))
	{
		newtScaleSet(s, ++progress);
		newtRefresh();	
		fprintf(flog, "%s", buffer);
	}
		
	pclose(p);
	rc = 0;
	
EXIT:
	newtFormDestroy(f);
	newtPopWindow();
		
	return rc;
}
